/* X-Chat
 * Copyright (C) 1998 Peter Zelezny.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include "../../config.h"
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <fcntl.h>
#include <ctype.h>
#include "../common/xchat.h"
#include "../common/cfgfiles.h"
#include "../common/userlist.h"
#include "simple.h"
#include <gdk/gdkkeysyms.h>
#include "gtkutil.h"
#include "menu.h"
#include "xtext.h"

extern GSList *fkey_list;
extern GtkWidget *main_window, *main_book;
extern int handle_multiline (struct session *sess, char *cmd, int history,
			     int nocommand);

extern void PrintText (struct session *sess, unsigned char *text);
extern int buf_get_line (char *, char **, int *, int len);

extern GSList *sess_list;
extern GSList *command_list;
extern GSList *replace_list;
extern GdkColor colors[];
extern GdkFont *font_normal;
extern GdkFont *dialog_font_normal;
extern GdkPixmap *channelwin_pix;
extern struct session *current_tab;
extern struct commands cmds[1];
extern struct xchatprefs prefs;

static int tab_nick_comp (GtkWidget * t, int shift);
static void nick_comp_chng (GtkWidget * t, int updown);
static void replace_handle (GtkWidget * wid);

static int key_action_delete_backward_word (GtkWidget * wid, GdkEventKey * evt,
					    char *d1, char *d2,
					    struct session *sess);
static int key_action_delete_forward_word (GtkWidget * wid, GdkEventKey * evt,
					   char *d1, char *d2,
					   struct session *sess);
static int key_action_handle_command (GtkWidget * wid, GdkEventKey * evt,
				      char *d1, char *d2, struct session *sess);
static int key_action_insert (GtkWidget * wid, 
			      GdkEventKey * evt, 
			      char *d1, 
			      char *d2,
			      struct session *sess);
static int key_action_scroll_page (GtkWidget * wid, GdkEventKey * evt, char *d1,
				   char *d2, struct session *sess);
static int key_action_set_buffer (GtkWidget * wid, GdkEventKey * evt, char *d1,
				  char *d2, struct session *sess);
static int key_action_history_up (GtkWidget * wid, GdkEventKey * evt, char *d1,
				  char *d2, struct session *sess);
static int key_action_history_down (GtkWidget * wid, GdkEventKey * evt,
				    char *d1, char *d2, struct session *sess);
static int key_action_tab_comp (GtkWidget * wid, GdkEventKey * evt, char *d1,
				char *d2, struct session *sess);
static int key_action_comp_chng (GtkWidget * wid, GdkEventKey * evt, char *d1,
				 char *d2, struct session *sess);
static int key_action_replace (GtkWidget * wid, GdkEventKey * evt, char *d1,
			       char *d2, struct session *sess);
static int key_action_put_history (GtkWidget * wid, GdkEventKey * evt, char *d1,
				   char *d2, struct session *sess);

static int key_action_buy_the_farm (GtkWidget * wid, 
				    GdkEventKey * evt, 
				    char *d1,
				    char *d2, 
				    struct session *sess);
static int key_action_close_window (GtkWidget * wid, 
				    GdkEventKey * evt, 
				    char *d1,
				    char *d2, 
				    struct session *sess);

typedef int (*HandlerFn) (GtkWidget * wid, 
			  GdkEventKey * evt, 
			  char *d1, 
			  char *d2,
			  struct session * sess);

struct key_binding {
	int mod;		/* Flags of STATE_* above */
	int keyval;		/* GDK keynumber */
	HandlerFn handler;     	/* Index into key_actions */
	char *data1, *data2;	/* Pointers to strings, these must be freed */
	struct key_binding *next;
} keys_root[] = {
	/*
	{ GDK_CONTROL_MASK, GDK_o,      key_action_insert, "%O", NULL },
	{ GDK_CONTROL_MASK, GDK_b,      key_action_insert, "%B", NULL },
	{ GDK_CONTROL_MASK, GDK_k,      key_action_insert, "%C", NULL },
	*/
	{ GDK_CONTROL_MASK, GDK_q,      key_action_buy_the_farm, NULL, NULL },
	{ GDK_CONTROL_MASK, GDK_w,      key_action_close_window, NULL, NULL },
	{ GDK_SHIFT_MASK,   GDK_Next,   key_action_comp_chng, NULL, NULL },
	{ GDK_SHIFT_MASK,   GDK_Prior,  key_action_comp_chng, NULL, NULL },
	{ 0,                GDK_Next,   key_action_scroll_page, NULL, NULL },
	{ 0,                GDK_Prior,  key_action_scroll_page, "Up", NULL },
	{ 0,                GDK_Down,   key_action_history_down, NULL, NULL },
	{ 0,                GDK_Up,     key_action_history_up, NULL, NULL },
	{ 0,                GDK_Tab,    key_action_tab_comp, NULL, NULL },
	{ 0,                GDK_space,  key_action_replace, NULL, NULL },
	{ 0,                GDK_Return, key_action_replace, NULL, NULL },
	{ GDK_CONTROL_MASK, GDK_Tab,    key_action_tab_comp, "Up", NULL },
	{ 0,                0,          NULL }
};

void
key_handle_send (GtkWidget * ignored, struct session *sess)
{
	char *cmd = gtk_entry_get_text (GTK_ENTRY (sess->gui->inputgad));
	if (cmd && cmd[0]) {
		handle_multiline (sess, cmd, TRUE, FALSE);
		gtk_entry_set_text (GTK_ENTRY (sess->gui->inputgad), "");
	}
}

int
key_handle_key_press (GtkWidget * wid, GdkEventKey * evt, struct session *sess)
{
	struct key_binding *kb = keys_root;
	int keyval = evt->keyval;
	int mod, n;

	mod = evt->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK | GDK_SHIFT_MASK);

	while (kb->handler) {
		if (kb->keyval == keyval && kb->mod == mod) {
			/* Run the function */
			n = kb->handler (wid, evt, kb->data1, kb->data2, sess);

			switch (n) {
			case 0:
				return 1;
			case 2:
				gtk_signal_emit_stop_by_name (
					GTK_OBJECT (wid),
					"key_press_event");
				return 1;
			}
		}

		kb++;
	}

	/* check if it's a return or enter */
	if ((evt->keyval == GDK_Return) || (evt->keyval == GDK_KP_Enter))
		key_handle_send (wid, sess);

	if (evt->keyval == GDK_KP_Enter)
		gtk_signal_emit_stop_by_name (GTK_OBJECT (wid),
					      "key_press_event");
	return 1;
}

/* ***** Key actions start here *********** */

/* See the NOTES above --AGL */

int
key_action_delete_forward_word (GtkWidget * wid, GdkEventKey * ent, char *d1,
				char *d2, struct session *sess)
{
	char *text, *cp;
	int len, curpos;
	char buf[16384];

	curpos = GTK_EDITABLE (wid)->current_pos;

	text = gtk_entry_get_text (GTK_ENTRY (wid));
	len = strlen (text);

	if (curpos >= len)
		return 2;

	if (len == 0 || len >= 16384)
		/* Not worth handling this case. */
		return 2;

	memmove (buf, text, curpos);
	buf[curpos] = 0;
	cp = &text[curpos];

	/* Skip leading whitespace. */
	while (isspace (*cp))
		++cp;

	/* Skip non-whitespace (i.e., a word). */
	while (*cp && !isspace (*cp))
		++cp;

	if (*cp)
		strcat (buf, cp);

	gtk_entry_set_text (GTK_ENTRY (wid), buf);
	GTK_EDITABLE (wid)->current_pos = curpos;
	return 2;
}

int
key_action_delete_backward_word (GtkWidget * wid, GdkEventKey * ent, char *d1,
				 char *d2, struct session *sess)
{
	char *text, *cp;
	int len, curpos;
	char buf[16384];

	curpos = GTK_EDITABLE (wid)->current_pos;

	if (curpos == 0)
		return 2;

	text = gtk_entry_get_text (GTK_ENTRY (wid));
	len = strlen (text);

	if (len == 0 || len >= 16384)
		/* Not worth handling this case. */
		return 2;

	memmove (buf, text, curpos);
	buf[curpos] = 0;
	cp = &buf[curpos - 1];

	/* Delete trailing whitespace. */
	while (cp >= buf && isspace (*cp))
		*cp-- = 0;

	/* Delete trailing non-whitespace (i.e., a word). */
	while (cp >= buf && !isspace (*cp))
		*cp-- = 0;

	if (curpos < len)
		strcat (buf, &text[curpos]);

	gtk_entry_set_text (GTK_ENTRY (wid), buf);
	GTK_EDITABLE (wid)->current_pos = curpos - (len - strlen (buf));
	return 2;
}

/* "Run command" */
static int
key_action_handle_command (GtkWidget * wid, GdkEventKey * evt, char *d1,
			   char *d2, struct session *sess)
{
	int ii, oi, len;
	char out[2048], d = 0;

	if (!d1)
		return 0;

	len = strlen (d1);

	/* Replace each "\n" substring with '\n' */
	for (ii = oi = 0; ii < len; ii++) {
		d = d1[ii];
		if (d == '\\') {
			ii++;
			d = d1[ii];
			if (d == 'n')
				out[oi++] = '\n';
			else if (d == '\\')
				out[oi++] = '\\';
			else {
				out[oi++] = '\\';
				out[oi++] = d;
			}
			continue;
		}
		out[oi++] = d;
	}
	out[oi] = 0;

	handle_multiline (sess, out, 0, 0);
	return 0;
}

int
key_action_insert (GtkWidget * wid, GdkEventKey * evt, char *d1, char *d2,
		   struct session *sess)
{
	int tmp_pos;

	if (!d1)
		return 1;

	tmp_pos = GTK_EDITABLE (wid)->current_pos;
	gtk_editable_insert_text (GTK_EDITABLE (wid), d1, strlen (d1),
				  &tmp_pos);
	GTK_EDITABLE (wid)->current_pos = tmp_pos;
	return 2;
}

/* handles PageUp/Down keys */
static int
key_action_scroll_page (GtkWidget * wid, GdkEventKey * evt, char *d1,
			char *d2, struct session *sess)
{
	int value, end;
	GtkAdjustment *adj;
	int up = 0;

	if (d1 && d1[0] != 0)
		up++;

	if (sess) {
		adj = GTK_RANGE (sess->gui->vscrollbar)->adjustment;
		if (up) {	/* PageUp */
			value = adj->value - adj->page_size;
			if (value < 0)
				value = 0;
		} else {	/* PageDown */
			end = adj->upper - adj->lower - adj->page_size;
			value = adj->value + adj->page_size;
			if (value > end)
				value = end;
		}
		gtk_adjustment_set_value (adj, value);
	}
	return 0;
}

static int
key_action_set_buffer (GtkWidget * wid, GdkEventKey * evt, char *d1, char *d2,
		       struct session *sess)
{
	if (!d1)
		return 1;
	if (d1[0] == 0)
		return 1;

	gtk_entry_set_text (GTK_ENTRY (wid), d1);
	return 2;
}

static int
key_action_history_up (GtkWidget * wid, GdkEventKey * ent, char *d1, char *d2,
		       struct session *sess)
{
	char *new_line;

	new_line = history_up (&sess->history);
	if (new_line)
		gtk_entry_set_text (GTK_ENTRY (wid), new_line);

	return 2;
}

static int
key_action_history_down (GtkWidget * wid, GdkEventKey * ent, char *d1,
			 char *d2, struct session *sess)
{
	char *new_line;

	new_line = history_down (&sess->history);
	if (new_line)
		gtk_entry_set_text (GTK_ENTRY (wid), new_line);

	return 2;
}

static int
key_action_tab_comp (GtkWidget * wid, GdkEventKey * ent, char *d1, char *d2,
		     struct session *sess)
{
	if (d1 && d1[0]) {
		if (tab_nick_comp (wid, 1) == -1)
			return 1;
		else
			return 2;
	} else {
		if (tab_nick_comp (wid, 0) == -1)
			return 1;
		else
			return 2;
	}
}

static int
key_action_comp_chng (GtkWidget * wid, GdkEventKey * ent, char *d1, char *d2,
		      struct session *sess)
{
	if (d1 && d1[0] != 0)
		nick_comp_chng (wid, 1);
	else
		nick_comp_chng (wid, 0);

	return 2;
}

static int
key_action_replace (GtkWidget * wid, GdkEventKey * ent, char *d1, char *d2,
		    struct session *sess)
{
	replace_handle (wid);
	return 1;
}

static int
key_action_put_history (GtkWidget * wid, GdkEventKey * ent, char *d1,
			char *d2, struct session *sess)
{
	history_add (&sess->history, gtk_entry_get_text (GTK_ENTRY (wid)));
	gtk_entry_set_text (GTK_ENTRY (wid), "");
	return 2;		/* -''- */
}

static int key_action_close_window (GtkWidget * wid, 
				    GdkEventKey * evt, 
				    char *d1,
				    char *d2, 
				    struct session *sess)
{
	gtk_widget_destroy (GTK_WIDGET (sess->gui->window));
	return 0;
}

static int key_action_buy_the_farm (GtkWidget * wid, 
				    GdkEventKey * evt, 
				    char *d1,
				    char *d2, 
				    struct session *sess)
{
	gtk_main_quit ();
	return 0;
}


/* -------- */


#define STATE_SHIFT	GDK_SHIFT_MASK
#define STATE_ALT	GDK_MOD1_MASK
#define STATE_CTRL	GDK_CONTROL_MASK

static void replace_handle (GtkWidget * t)
{
	char *text, *postfix_pnt;
	struct popup *pop;
	GSList *list = replace_list;
	char word[256];
	char postfix[256];
	char outbuf[4096];
	int c, len, xlen;

	text = gtk_entry_get_text (GTK_ENTRY (t));

	len = strlen (text);
	for (c = len - 1; c > 0; c--) {
		if (text[c] == ' ')
			break;
	}
	if (text[c] == ' ')
		c++;
	xlen = c;
	if (len - c >= (sizeof (word) - 12))
		return;
	if (len - c < 1)
		return;
	memcpy (word, &text[c], len - c);
	word[len - c] = 0;
	len = strlen (word);
	if (word[0] == '\'' && word[len] == '\'')
		return;
	postfix_pnt = NULL;
	for (c = 0; c < len; c++) {
		if (word[c] == '\'') {
			postfix_pnt = &word[c + 1];
			word[c] = 0;
			break;
		}
	}

	if (postfix_pnt != NULL) {
		if (strlen (postfix_pnt) > sizeof (postfix) - 12)
			return;
		strcpy (postfix, postfix_pnt);
	}
	while (list) {
		pop = (struct popup *) list->data;
		if (strcmp (pop->name, word) == 0) {
			memcpy (outbuf, text, xlen);
			outbuf[xlen] = 0;
			if (postfix_pnt == NULL)
				snprintf (word, sizeof (word), "%s", pop->cmd);
			else
				snprintf (word, sizeof (word), "%s%s", pop->cmd,
					  postfix);
			strcat (outbuf, word);
			gtk_entry_set_text (GTK_ENTRY (t), outbuf);
			return;
		}
		list = list->next;
	}
}

static struct session *find_session_from_inputgad (GtkWidget * w)
{
	GSList *list = sess_list;
	struct session *sess;

	/* First find the session from the widget */
	while (list) {
		sess = (struct session *) list->data;
		if (sess->gui->inputgad == w)
			return sess;
		list = list->next;
	}

	/*Erm, we didn't find ANY valid session, HELP! */
	return NULL;
}

static int nick_comp_get_nick (char *tx, char *n)
{
	int c, len = strlen (tx);

	for (c = 0; c < len; c++) {
		if (tx[c] == ':' || tx[c] == ','
		    || tx[c] == prefs.nick_suffix[0]) {
			n[c] = 0;
			return 0;
		}
		if (tx[c] == ' ' || tx[c] == '.' || tx[c] == 0)
			return -1;
		n[c] = tx[c];
	}
	return -1;
}

static void nick_comp_chng (GtkWidget * t, int updown)
{
	struct session *sess;
	struct User *user, *last = NULL;
	char *text, nick[64];
	int len, slen;
	GSList *list;

	sess = find_session_from_inputgad (t);
	if (sess == NULL)
		return;

	text = gtk_entry_get_text (GTK_ENTRY (t));

	if (nick_comp_get_nick (text, nick) == -1)
		return;
	len = strlen (nick);

	list = sess->userlist;
	while (list) {
		user = (struct User *) list->data;
		slen = strlen (user->nick);
		if (len != slen) {
			last = user;
			list = list->next;
			continue;
		}
		if (strncasecmp (user->nick, nick, len) == 0) {
			if (updown == 0) {
				if (list->next == NULL)
					return;
				user->weight--;
				((struct User *) list->next->data)->weight++;
				snprintf (nick, sizeof (nick), "%s%c ",
					  ((struct User *) list->next->data)->
					  nick, prefs.nick_suffix[0]);
			} else {
				if (last == NULL)
					return;
				user->weight--;
				last->weight++;
				snprintf (nick, sizeof (nick), "%s%c ",
					  last->nick, prefs.nick_suffix[0]);
			}
			gtk_entry_set_text (GTK_ENTRY (t), nick);
			return;
		}
		last = user;
		list = list->next;
	}
}

static void tab_comp_find_common (char *a, char *b)
{
	int c;
	int alen = strlen (a), blen = strlen (b), len;

	if (blen > alen)
		len = alen;
	else
		len = blen;
	for (c = 0; c < len; c++) {
		if (a[c] != b[c]) {
			a[c] = 0;
			return;
		}
	}
	a[c] = 0;
}

static void tab_comp_cmd (GtkWidget * t)
{
	char *text, *last = NULL, *cmd, *postfix = NULL;
	GSList *list = command_list;
	struct popup *pop;
	int len, i, slen;
	struct session *sess;
	char buf[2048];
	char lcmd[2048];

	text = gtk_entry_get_text (GTK_ENTRY (t));

	sess = find_session_from_inputgad (t);

	text++;
	len = strlen (text);
	if (len == 0)
		return;
	for (i = 0; i < len; i++) {
		if (text[i] == ' ') {
			postfix = &text[i + 1];
			text[i] = 0;
			len = strlen (text);
			break;
		}
	}

	while (list) {
		pop = (struct popup *) list->data;
		slen = strlen (pop->name);
		if (len > slen) {
			list = list->next;
			continue;
		}
		if (strncasecmp (pop->name, text, len) == 0) {
			if (last == NULL) {
				last = pop->name;
				snprintf (lcmd, sizeof (lcmd), "%s", last);
			} else if (last > (char *) 1) {
				snprintf (buf, sizeof (buf), "%s %s ", last,
					  pop->name);
				PrintText (sess, buf);
				/*PrintText (sess, last);
				   PrintText (sess, "\t");
				   PrintText (sess, pop->name);
				   PrintText (sess, "\t"); */
				last = (void *) 1;
				tab_comp_find_common (lcmd, pop->name);
			} else if (last == (void *) 1) {
				PrintText (sess, pop->name);
				/*PrintText (sess, "\t"); */
				tab_comp_find_common (lcmd, pop->name);
			}
		}
		list = list->next;
	}

	i = 0;
	while (cmds[i].name != NULL) {
		cmd = cmds[i].name;
		slen = strlen (cmd);
		if (len > slen) {
			i++;
			continue;
		}
		if (strncasecmp (cmd, text, len) == 0) {
			if (last == NULL) {
				last = cmd;
				snprintf (lcmd, sizeof (lcmd), "%s", last);
			} else if (last > (char *) 1) {
				snprintf (buf, sizeof (buf), "%s %s ", last,
					  cmd);
				PrintText (sess, buf);
				/*PrintText (sess, last);
				   PrintText (sess, "\t");
				   PrintText (sess, cmd);
				   PrintText (sess, "\t"); */
				last = (void *) 1;
				tab_comp_find_common (lcmd, cmd);
			} else if (last == (void *) 1) {
				PrintText (sess, cmd);
				/*PrintText (sess, cmd);
				   PrintText (sess, "\t"); */
				tab_comp_find_common (lcmd, cmd);
			}
		}
		i++;
	}

	if (last == NULL)
		return;
	if (last == (void *) 1)
		PrintText (sess, "\n");

	if (last > (char *) 1) {
		if (strlen (last) > (sizeof (buf) - 1))
			return;
		if (postfix == NULL)
			snprintf (buf, sizeof (buf), "/%s ", last);
		else
			snprintf (buf, sizeof (buf), "/%s %s", last, postfix);
		gtk_entry_set_text (GTK_ENTRY (t), buf);
		return;
	} else if (strlen (lcmd) > (sizeof (buf) - 1))
		return;
	if (postfix == NULL)
		snprintf (buf, sizeof (buf), "/%s", lcmd);
	else
		snprintf (buf, sizeof (buf), "/%s %s", lcmd, postfix);
	gtk_entry_set_text (GTK_ENTRY (t), buf);
}

/* In the following 'b4' is *before* the text (just say b4 and before out loud)
   and c5 is *after* (because c5 is next after b4, get it??) --AGL */

static int
tab_nick_comp_next (struct session *sess, GtkWidget * wid, char *b4,
		    char *nick, char *c5, int shift)
{
	struct User *user = 0, *last = NULL;
	char buf[4096];
	GSList *list;

	list = sess->userlist;
	while (list) {
		user = (struct User *) list->data;
		if (strcmp (user->nick, nick) == 0)
			break;
		last = user;
		list = list->next;
	}
	if (!list)
		return 0;
	if (shift) {
		if (last)
			snprintf (buf, 4096, "%s %s%s", b4, last->nick, c5);
		else
			snprintf (buf, 4096, "%s %s%s", b4, nick, c5);
	} else {
		if (list->next)
			snprintf (buf, 4096, "%s %s%s", b4,
				  ((struct User *) list->next->data)->nick, c5);
		else {
			if (sess->userlist)
				snprintf (buf, 4096, "%s %s%s", b4,
					  ((struct User *) sess->userlist->
					   data)->nick, c5);
			else
				snprintf (buf, 4096, "%s %s%s", b4, nick, c5);
		}
	}
	gtk_entry_set_text (GTK_ENTRY (wid), buf);

	return 1;
}

static int tab_nick_comp (GtkWidget * t, int shift)
{
	struct session *sess;
	struct User *user = NULL, *match_user = NULL;
	char *text, not_nick_chars[16] = "";
	int len, slen, first = 0, i, j, match_count = 0, match_pos = 0;
	char buf[2048], nick_buf[2048] = { 0 }, *b4 = NULL, *c5 =
	    NULL, *match_text = NULL, match_char = -1, *nick =
	    NULL, *current_nick = NULL;
	GSList *list = NULL, *match_list = NULL, *first_match = NULL;

	text = gtk_entry_get_text (GTK_ENTRY (t));

	sess = find_session_from_inputgad (t);
	if (sess == NULL)
		return 0;
	if (sess->is_dialog)
		return 0;

	len = strlen (text);
	if (!strchr (text, ' ')) {
		if (text[0] == '/') {
			tab_comp_cmd (t);
			return 0;
		}
	}

	/* Is the text more than just a nick? */

	sprintf (not_nick_chars, " .?%c", prefs.nick_suffix[0]);

	if (strcspn (text, not_nick_chars) != strlen (text)) {
		/* If we're doing old-style nick completion and the text input widget
		 * contains a string of the format: "nicknameSUFFIX" or "nicknameSUFFIX ",
		 * where SUFFIX is the Nickname Completion Suffix character, then cycle
		 * through the available nicknames.
		 */
		if (prefs.old_nickcompletion) {
			char *space = strchr (text, ' ');

			if ((!space || space == &text[len - 1]) &&
			    text[len - (space ? 2 : 1)] == prefs.nick_suffix[0]) {
				/* This causes the nickname to cycle. */
				nick_comp_chng (t, shift);
				return 0;
			}
		}

		j = GTK_EDITABLE (t)->current_pos;

		/* !! FIXME !! */
		if (len - j < 0)
			return 0;

		b4 = (char *) malloc (len + 1);
		c5 = (char *) malloc (len + 1);
		memmove (c5, &text[j], len - j);
		c5[len - j] = 0;
		memcpy (b4, text, len + 1);

		for (i = j - 1; i > -1; i--) {
			if (b4[i] == ' ') {
				b4[i] = 0;
				break;
			}
			b4[i] = 0;
		}
		memmove (text, &text[i + 1], (j - i) + 1);
		text[(j - i) - 1] = 0;

		if (tab_nick_comp_next (sess, t, b4, text, c5, shift)) {
			free (b4);
			free (c5);
			return 1;
		}
		first = 0;
	} else
		first = 1;

	len = strlen (text);

	if (text[0] == 0)
		return -1;

	list = sess->userlist;

	/* make a list of matches */
	while (list) {
		user = (struct User *) list->data;
		slen = strlen (user->nick);
		if (len > slen) {
			list = list->next;
			continue;
		}
		if (strncasecmp (user->nick, text, len) == 0)
			match_list = g_slist_prepend (match_list, user->nick);
		list = list->next;
	}
	match_list = g_slist_reverse (match_list);	/* faster then _append */
	match_count = g_slist_length (match_list);

	/* no matches, return */
	if (match_count == 0) {
		if (!first) {
			free (b4);
			free (c5);
		}
		return 0;
	}
	first_match = match_list;
	match_pos = len;

	/* if we have more then 1 match, we want to act like readline and grab common chars */
	if (!prefs.old_nickcompletion && match_count > 1) {
		while (1) {
			while (match_list) {
				current_nick =
				    (char *)
				    malloc (strlen ((char *) match_list->data) +
					    1);
				strcpy (current_nick,
					(char *) match_list->data);
				if (match_char == -1) {
					match_char = current_nick[match_pos];
					match_list = g_slist_next (match_list);
					free (current_nick);
					continue;
				}
				if (tolower (current_nick[match_pos]) !=
				    tolower (match_char)) {
					match_text =
					    (char *) malloc (match_pos);
					current_nick[match_pos] = '\0';
					strcpy (match_text, current_nick);
					free (current_nick);
					match_pos = -1;
					break;
				}
				match_list = g_slist_next (match_list);
				free (current_nick);
			}

			if (match_pos == -1)
				break;

			match_list = first_match;
			match_char = -1;
			++match_pos;
		}

		match_list = first_match;
	} else
		match_user = (struct User *) match_list->data;

	/* no match, if we found more common chars among matches, display them in entry */
	if (match_user == NULL) {
		while (match_list) {
			nick = (char *) match_list->data;
			sprintf (nick_buf, "%s%s ", nick_buf, nick);
			match_list = g_slist_next (match_list);
		}
		PrintText (sess, nick_buf);

		if (first)
			snprintf (buf, sizeof (buf), "%s", match_text);
		else {
			snprintf (buf, sizeof (buf), "%s %s%s", b4, match_text,
				  c5);
			GTK_EDITABLE (t)->current_pos =
			    strlen (b4) + strlen (match_text);
			free (b4);
			free (c5);
		}
		free (match_text);
	} else {
		if (first)
			snprintf (buf, sizeof (buf), "%s%c ", match_user->nick,
				  prefs.nick_suffix[0]);
		else {
			snprintf (buf, sizeof (buf), "%s %s%s", b4,
				  match_user->nick, c5);
			GTK_EDITABLE (t)->current_pos =
			    strlen (b4) + strlen (match_user->nick);
			free (b4);
			free (c5);
		}
	}
	gtk_entry_set_text (GTK_ENTRY (t), buf);

	return 0;
}
