/* X-Chat
 * Copyright (C) 1998 Peter Zelezny.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <stdio.h>
#include <string.h>
#include "../common/xchat.h"
#include "simple.h"
#include "gtkutil.h"
#include "xtext.h"


extern char *nocasestrstr (char *text, char *tofind);
extern int is_session (session * sess);

struct SearchData {
	GnomeDialog *dialog;
	GnomeEntry *gnome_entry;
	GtkEntry *entry;
	gint signal_id;

	session *sess;
};

static void
session_kill (GtkObject *object,
	      struct SearchData *data)
{
	gnome_dialog_close (data->dialog);
}

static void search_search (struct SearchData *data,
			   const char *text)
{
	session *sess;
	static void *last = NULL;

	sess = data->sess;

	/* This probably isn't needed anymore. */
	if (!is_session (sess)) {
		gtkutil_simpledialog ("The window you opened this Search "
				      "for doesn't exist anymore.");
		return;
	}
	
	last = gtk_xtext_search (GTK_XTEXT (sess->gui->textgad), 
				 (char *) text, last);
}

static void search_entry_cb (GtkWidget *entry, 
			     struct SearchData *data)
{
	search_search (data, gtk_entry_get_text (GTK_ENTRY (entry)));
}

static void
dialog_clicked_cb (GnomeDialog *dialog,
		   int button,
		   struct SearchData *data)
{
	char *text;

	switch (button) {
	case 0:
		text = gtk_entry_get_text (data->entry);
		gnome_entry_prepend_history (GNOME_ENTRY (data->gnome_entry), 
					     TRUE, text);
		search_search (data, text);
		break;

	default:
		gtk_signal_disconnect (GTK_OBJECT (data->sess->gui->window),
				       data->signal_id);
		gnome_dialog_close (dialog);
		break;
	}
}

static void
dialog_closed_cb (GtkObject *object,
		  struct SearchData *data)
{
	g_free (data);
}

void search_open (session * sess)
{
	struct SearchData *data;

	data = g_new (struct SearchData, 1);
	data->sess = sess;
	data->dialog = GNOME_DIALOG (gnome_dialog_new (_("Search"),
						       GNOME_STOCK_PIXMAP_SEARCH,
						       GNOME_STOCK_BUTTON_CLOSE, 
						       NULL));

	data->gnome_entry = GNOME_ENTRY (gnome_entry_new ("monkeytalk:search"));
	data->entry = GTK_ENTRY (gnome_entry_gtk_entry (data->gnome_entry));
	data->signal_id = gtk_signal_connect (GTK_OBJECT (sess->gui->window),
					      "destroy", 
					      GTK_SIGNAL_FUNC (session_kill),
					      data);

	gtk_signal_connect (GTK_OBJECT (data->entry), "activate",
			    GTK_SIGNAL_FUNC (search_entry_cb), data);

	gtk_box_pack_start (GTK_BOX (data->dialog->vbox),
			    GTK_WIDGET (data->gnome_entry), TRUE, TRUE, 0);
	gtk_signal_connect (GTK_OBJECT (data->dialog), "clicked",
			    GTK_SIGNAL_FUNC (dialog_clicked_cb), data);
	gtk_signal_connect (GTK_OBJECT (data->dialog), "destroy",
			    GTK_SIGNAL_FUNC (dialog_closed_cb), data);

	gtk_widget_show_all (GTK_WIDGET (data->dialog));
}
